from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Dict, Optional, Sequence, Union

from airflow.models.taskmixin import TaskMixin
from airflow.models.baseoperator import TaskStateChangeCallback
from airflow.sensors.base import BaseSensorOperator
from airflow.utils import timezone

ONE_WEEK = 60 * 60 * 24 * 7
FIVE_MINUTES = 60 * 5

SENSOR_TASK_ID = "wait_for_event"


class TriggerDagConfiguration(ABC):
    """
    This is the parameters for triggering other dags, based on a supplied sensor.

    :param trigger_dag_id: The id of the DAG to trigger on succesful sensor (Templated)
    :type trigger_dag_id: str
    :param dag_id: The id of the DAG. If None it is trigger_dag_id+_trigger
    :type dag_id: str
    :param poke_interval: Time in seconds that the job should wait in
        between each tries. Default is 5 minutes.
    :type poke_interval: int
    :param timeout: Time, in seconds before the sensor task times out and fails.
        Think of it as the time within we expect an event to happen.
        Default is one week.
    :type timeout: int
    :param mode: How the sensor operates. (Only for versions above 1.10.2)
        Options are: ``{ poke | reschedule }``, default is ``reschedule``.
        When set to ``poke`` the sensor is taking up a worker slot for its
        whole execution time and sleeps between pokes. Use this mode if the
        expected runtime of the sensor is short or if a short poke interval
        is required. Note that the sensor will hold onto a worker slot and
        a pool slot for the duration of the sensor's runtime in this mode.
        When set to ``reschedule`` the sensor task frees the worker slot when
        the criteria is not yet met and it's rescheduled at a later time. Use
        this mode if the time before the criteria is met is expected to be
        quite long. The poke interval should be more than one minute to
        prevent too much load on the scheduler.
    :type mode: str
    """

    def __init__(self,
                 trigger_dag_id: str,
                 dag_id: str = None,
                 mode: str = 'reschedule',
                 poke_interval: Optional[int] = FIVE_MINUTES,
                 timeout: Optional[int] = ONE_WEEK,
                 schedule_interval: timedelta = timedelta(minutes=10),
                 sensor_on_retry_callback: Optional[TaskStateChangeCallback] = None,
                 sensor_retries: Optional[int] = None,
                 clear_dag=False,
                 configuration: Dict = None,
                 ) -> None:
        self.trigger_dag_id = trigger_dag_id
        self.dag_id = dag_id if dag_id else f'{trigger_dag_id}_trigger'
        self.mode = mode
        self.poke_interval = poke_interval
        self.timeout = timeout
        self.schedule_interval = schedule_interval
        self.sensor_on_retry_callback = sensor_on_retry_callback
        self.sensor_retries = sensor_retries
        self.clear_dag = clear_dag
        self.configuration = configuration if configuration else {}
        # Update configuration with some sensible defaults
        self.configuration = {**self.configuration,
                              'trigger_run_id': '{{ run_id }}',
                              'trigger_dag': '{{ dag.dag_id }}',
                              'trigger_dag_task': '{{ task.task_id }}',
                              'trigger_dag_task_url': '{{ task_instance.log_url }}',
                              }

    def execution_date(*args, **kwargs):
        """
        Override this method to return the appropriate execution date
         for the DAG to trigger.

        Default is to evaluate to now.
        """
        return timezone.utcnow()

    @abstractmethod
    def create_sensor(self) -> BaseSensorOperator:
        """
        Override this and return the sensor that starts off the triggering
        """
        pass

    def create_downstream_sensor(self) -> Union[TaskMixin, Sequence[TaskMixin]]:
        """
        Override this and return tasks that should be downstreams of the sensor
        and upstreams of the triggering
        """
        pass

    def create_downstream_triggering(self) -> Union[TaskMixin, Sequence[TaskMixin]]:
        """
        Override this and return tasks that should be downstreams of triggering
        """
        pass

    def default_doc_md(self) -> str:
        """
        Override this and return a self aware doc_md.
        Good for default description of the pipeline.
        """
        pass



from airflow.contrib.operators.slack_webhook_operator import SlackWebhookOperator


def slack_message_on_context(http_conn_id: str = None,
                             webhook_token: str = None,
                             message: str = "",
                             attachments=None,
                             channel: str = None,
                             username: str = None,
                             icon_emoji: str = None,
                             icon_url: str = None,
                             link_names=False,
                             blocks=None,
                             proxy=None):
    """
    This is a wrapper function around the SlackWebhookOperator
     You can for example utilise it as a failure or retry function.
     Example:
     ...
     on_retry_callback=slack_message_on_context(message="Sensor timeout, we are doing another wait.")
     ...

    :param http_conn_id: connection that has Slack webhook token in the extra field
    :type http_conn_id: str
    :param webhook_token: Slack webhook token
    :type webhook_token: str
    :param message: The message you want to send on Slack
    :type message: str
    :param attachments: The attachments to send on Slack. Should be a list of
                        dictionaries representing Slack attachments.
    :type attachments: list
    :param channel: The channel the message should be posted to
    :type channel: str
    :param username: The username to post to slack with
    :type username: str
    :param icon_emoji: The emoji to use as icon for the user posting to Slack
    :type icon_emoji: str
    :param icon_url: The icon image URL string to use in place of the default icon.
    :type icon_url: str
    :param link_names: Whether or not to find and link channel and usernames in your
                       message
    :type link_names: bool
    :param blocks: The blocks to send on Slack. Should be a list of
        dictionaries representing Slack blocks. (Only Airflow > 1.10.7)
    :type blocks: list
    :param proxy: Proxy to use to make the Slack webhook call
    :type proxy: str
    :return: a function that sends a Slack message using the SlackWebhookOperator
    """

    def on_context(context):
        retry_warning = SlackWebhookOperator(
            task_id='slack_failed_alert',
            http_conn_id=http_conn_id,
            webhook_token=webhook_token,
            message=message.format(**context),
            username=username,
            icon_emoji=icon_emoji,
            icon_url=icon_url,
            link_names=link_names,
            proxy=proxy,
            channel=channel,
            blocks=blocks,
            attachments=attachments
        )

        return retry_warning.execute(context)

    return on_context



from airflow import DAG
from airflow.operators.dagrun_operator import TriggerDagRunOperator
from airflow.operators.python_operator import PythonOperator
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.task_group import TaskGroup
from unaflow.dags.trigger_dag.trigger_dag_configuration import \
    TriggerDagConfiguration


def trigger_dag_factory(config: TriggerDagConfiguration, **dag_kwargs):
    """
    A factory for creating a DAG that is able to listen for events, from
    a supplied sensor in the config.

    The factory ensures that we only trigger an external DAG once per event.

    It basically follows this DAG graph

                     -------------------------------------------
    wait_for_event-->| evaluate_execution_date --> trigger_dag |
                     -------------------------------------------

    You are able to override the evaluate_execution_date function, with the
     supplied config.
    You are also able to add extra tasks betweem wait_for_event and
     evaluate_execution_date, by overriding the sensor_downstreams function
    in the config. And you can add tasks after trigger_dag, by overriding the
    triggering_downstreams function in the config.

    param: config: The configuration to be used. The config needs to at least
    supply a sensor.
    type: config: TriggerDagConfiguration
    param: dag_kwargs: These are the extra arguments to set on the dag. Certain
     values are not able to be set, such as max_active_runs, catchup,
    dag_id abd schedule interval (set in config)
    """

    # Use default doc_md from config if nothing is set
    if 'doc_md' not in dag_kwargs:
        dag_kwargs['doc_md'] = config.default_doc_md()

    with DAG(
            dag_id=config.dag_id,
            catchup=False,
            max_active_runs=1,
            schedule_interval=config.schedule_interval,
            **dag_kwargs) as dag:
        # Create the START sensor task
        # Subclasses needs to implement the _create_sensor method
        # with the appropriate sensor
        sensor: BaseSensorOperator = config.create_sensor()
        sensor.wait_for_downstream = True
        sensor.depends_on_past = True  # Need to set this as well as wait_for_downstream
        if config.sensor_on_retry_callback:
            sensor.on_retry_callback = config.sensor_on_retry_callback
        if config.sensor_retries:
            sensor.retries = config.sensor_retries

        with TaskGroup(group_id="trigger_dag") as tg:
            # Get the timestamp to trigger. Either it is default, now,
            # or supplied from the callable trigger_dag_execution_callable
            trigger_dag_execution_date = PythonOperator(
                task_id="evaluate_execution_date",
                python_callable=config.execution_date,
                provide_context=True,
            )

            # Trigger the external DAG
            # Merge DAG configuration from Configuration
            # with some sensible defaults
            # This is the DAG conf that is by default sent to the
            trigger_dag = TriggerDagRunOperator(
                task_id=f"trigger_{config.trigger_dag_id}",
                trigger_dag_id=config.trigger_dag_id,
                conf=config.configuration,
                execution_date='{{ ti.xcom_pull(task_ids="%s") }}' % trigger_dag_execution_date.task_id,
                reset_dag_run=config.clear_dag,
            )

            trigger_dag_execution_date >> trigger_dag

        sensor >> tg

        # Set downstreams to the sensor, if given.
        # Also make sure they are added as upstreams
        # to triggering
        sensor_downstreams = config.create_downstream_sensor()
        if sensor_downstreams:
            sensor.set_downstream(sensor_downstreams)
            tg.set_upstream(sensor_downstreams)

        # Set downstreams to the triggering, if given.
        triggering_downstreams = config.create_downstream_triggering()
        if triggering_downstreams:
            tg.set_downstream(triggering_downstreams)

    return dag



import re

from airflow.models import TaskInstance
from unaflow.dags.trigger_dag.trigger_dag_configuration import SENSOR_TASK_ID
from typing import Pattern
from airflow.utils import timezone


def find_first_file(ti: TaskInstance):
    """
    Finds the first result in the sorted xcom result of the sensor task.
    """
    # Get the files that triggered the sensor
    file_list = ti.xcom_pull(task_ids=SENSOR_TASK_ID)
    # Sort the files. Just to make some sense of out things
    sorted_file_list = sorted(file_list)
    # Return the top file to the xcom_of this task_instance
    return sorted_file_list[0]


def evaluate_execution_date_on_file_suffix(ti: TaskInstance, pattern: Pattern = re.compile('.*_([0-9]{8})')):
    """
    We utilise find_first_file and get the filename.
    From that filename, we derive the executiondate based on the suffix
    and the given pattern of the filename
    """

    filename = find_first_file(ti)
    match = pattern.match(filename)
    return str(timezone.parse(match.group(1), timezone=timezone.utc))



from setuptools import setup, find_packages
import versioneer

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='unaflow',
    url='https://github.com/unacast/unaflow',
    author='Unacast',
    author_email='ole.christian.langfjaran@unacast.com',
    description='A collection helper classes for Apache Airflow',
    license='Apache License 2.0',
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[],  # Meant as a helper library. User is required to install.
    python_requires='>=3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
